@@ -20,6 +20,7 @@ import collections
 import fcntl
 import hashlib
 import hmac
+import operator
 import os
 import random
 import select
@@ -43,27 +44,44 @@ iothread = None  # the thread in which all IO will be performed
                  # This thread will tuck away the threading situation such that
                  # calling code doesn't have to do any gymnastics to cope with
                  # the nature of things.
+iothreadready = False  # whether io thread is yet ready to work
+iothreadwaiters = []  # threads waiting for iothreadready
+ignoresockets = set()  # between 'select' firing and 'recvfrom', a socket
+                       # should be ignored
 ioqueue = collections.deque([])
 selectbreak = None
 selectdeadline = 0
 running = True
 iosockets = []  # set of iosockets that will be shared amongst Session objects
+MAX_BMCS_PER_SOCKET = 64  # no more than this many BMCs will share a socket
+                         # this could be adjusted based on rmem_max
+                         # value, leading to fewer filehandles
 
 
-def _ioworker(initialized):
+def _ioworker():
+    global iothreadready
     global selectbreak
     global selectdeadline
     selectbreak = os.pipe()
     fcntl.fcntl(selectbreak[0], fcntl.F_SETFL, os.O_NONBLOCK)
-    iosockets.append(selectbreak[0])
     iowaiters = []
     timeout = 300
-    initialized.set()
+    iothreadready = True
     while running:
+        while iothreadwaiters:
+            waiter = iothreadwaiters.pop()
+            waiter.set()
         if timeout < 0:
             timeout = 0
         selectdeadline = _monotonic_time() + timeout
-        tmplist, _, _ = select.select(iosockets, (), (), timeout)
+        if ignoresockets:
+            mysockets = [selectbreak[0]]
+            for pendingsocket in iosockets:
+                if pendingsocket not in ignoresockets:
+                    mysockets.append(pendingsocket)
+        else:
+            mysockets = iosockets + [selectbreak[0]]
+        tmplist, _, _ = select.select(mysockets, (), (), timeout)
         # pessimistically move out the deadline
         # doing it this early (before ioqueue is evaluated)
         # this avoids other threads making a bad assumption
@@ -80,6 +98,7 @@ def _ioworker(initialized):
                     # was the endgame
                     pass
             else:
+                ignoresockets.add(handle)
                 rdylist.append(handle)
         for w in iowaiters:
             w[2].append(tuple(rdylist))
@@ -117,10 +136,6 @@ def _io_apply(function, args):
     return result[0]
 
 
-selectdeadline = 0
-selectwait = None
-
-
 def _io_sendto(mysocket, packet, sockaddr):
     #Want sendto to act reasonably sane..
     mysocket.setblocking(1)
@@ -129,6 +144,7 @@ def _io_sendto(mysocket, packet, sockaddr):
 
 def _io_recvfrom(mysocket, size):
     mysocket.setblocking(0)
+    ignoresockets.discard(mysocket)
     try:
         return mysocket.recvfrom(size)
     except socket.error:
@@ -147,7 +163,7 @@ def _monotonic_time():
     return os.times()[4]
 
 
-def _poller(readhandles, timeout=0):
+def _poller(timeout=0):
     rdylist = _io_apply('wait', timeout + _monotonic_time())
     return rdylist
 
@@ -189,6 +205,14 @@ def get_ipmi_error(response, suffix=""):
     return res
 
 
+def _checksum(*data):  # Two's complement over the data
+    csum = sum(data)
+    csum ^= 0xff
+    csum += 1
+    csum &= 0xff
+    return csum
+
+
 class Session(object):
     """A class to manage common IPMI session logistics
 
@@ -215,8 +239,12 @@ class Session(object):
     keepalive_sessions = {}
     peeraddr_to_nodes = {}
     iterwaiters = []
-    # Upon exit of python, make sure we play nice with BMCs by assuring closed
-    # sessions for all that we tracked
+    #NOTE(jbjohnso):
+    #socketpool is a mapping of sockets to usage count
+    socketpool = {}
+    #this will be a lock.  Delay the assignment so that a calling framework
+    #can do something like reassign our threading and select modules
+    socketchecking = None
 
     @classmethod
     def _cleanup(cls):
@@ -227,52 +255,40 @@ class Session(object):
         running = False
 
     @classmethod
-    def _createsocket(cls):
-        global iowork
+    def _initsessions(cls):
+        atexit.register(cls._cleanup)
+
+    def _assignsocket(self):
         global iothread
+        global iothreadready
         global iosockets
         if iothread is None:
+            self._initsessions()
             initevt = threading.Event()
-            iothread = threading.Thread(target=_ioworker, args=(initevt,))
+            iothreadwaiters.append(initevt)
+            iothread = threading.Thread(target=_ioworker)
             iothread.daemon = True
             iothread.start()
             initevt.wait()
-        atexit.register(cls._cleanup)
-        cls.socket = _io_apply(socket.socket,
-                               (socket.AF_INET6, socket.SOCK_DGRAM))  # INET6
+        elif not iothreadready:
+            initevt = threading.Event()
+            iothreadwaiters.append(initevt)
+            initevt.wait()
+        # seek for the least used socket.  As sessions close, they may free
+        # up slots in seemingly 'full' sockets.  This scheme allows those
+        # slots to be recycled
+        sorted_candidates = sorted(self.socketpool.iteritems(),
+                                   key=operator.itemgetter(1))
+        if sorted_candidates and sorted_candidates[0][1] < MAX_BMCS_PER_SOCKET:
+            self.socketpool[sorted_candidates[0][0]] += 1
+            return sorted_candidates[0][0]
+        # we need a new socket
+        tmpsocket = _io_apply(socket.socket,
+                              (socket.AF_INET6, socket.SOCK_DGRAM))  # INET6
                                     # can do IPv4 if you are nice to it
-        try:  # we will try to fixup our receive buffer size if we are smaller
-             # than allowed.
-            maxmf = open("/proc/sys/net/core/rmem_max")
-            rmemmax = int(maxmf.read())
-            rmemmax = rmemmax / 2
-            curmax = cls.socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
-            curmax = curmax / 2
-            if (rmemmax > curmax):
-                _io_apply(cls.socket.setsockopt, (socket.SOL_SOCKET,
-                                                  socket.SO_RCVBUF,
-                                                  rmemmax))
-        except Exception:
-            # FIXME: be more selective in catching exceptions
-            pass
-
-        curmax = _io_apply(cls.socket.getsockopt,
-                           (socket.SOL_SOCKET, socket.SO_RCVBUF))
-        iosockets.append(cls.socket)
-        curmax = curmax / 2
-        # we throttle such that we never have no more outstanding packets than
-        # our receive buffer should be able to handle
-        cls.pending = 0
-        cls.maxpending = curmax / 1000
-        # pessimistically assume 1 kilobyte messages,
-        # which is way larger than almost all ipmi datagrams.
-        # For faster performance, sysadmins may want to examine and tune
-        # /proc/sys/net/core/rmem_max up.  This allows the module to request
-        # more, but does not increase buffers for applications that do less
-        # creative things
-        # TODO(jbjohnso): perhaps spread sessions across a socket pool when
-        # rmem_max is small, still get ~65/socket, but avoid long queues that
-        # might happen with low rmem_max and putting thousands of nodes in line
+        self.socketpool[tmpsocket] = 1
+        iosockets.append(tmpsocket)
+        return tmpsocket
 
     def _sync_login(self, response):
         """Handle synchronous callers in liue of
@@ -291,13 +307,14 @@ class Session(object):
         trueself = None
         for res in socket.getaddrinfo(bmc, port, 0, socket.SOCK_DGRAM):
             sockaddr = res[4]
-            if (res[0] == socket.AF_INET):  # convert the sockaddr to AF_INET6
+            if res[0] == socket.AF_INET:  # convert the sockaddr to AF_INET6
                 newhost = '::ffff:' + sockaddr[0]
                 sockaddr = (newhost, sockaddr[1], 0, 0)
             if sockaddr in cls.bmc_handlers:
                 self = cls.bmc_handlers[sockaddr]
                 if (self.bmc == bmc and self.userid == userid and
-                        self.password == password and self.kgo == kg):
+                        self.password == password and self.kgo == kg and
+                        self.logged):
                     trueself = self
                 else:
                     del cls.bmc_handlers[sockaddr]
@@ -323,12 +340,15 @@ class Session(object):
                 else:
                     self.iterwaiters.append(onlogon)
             return
+        self.privlevel = 4
         self.maxtimeout = 3  # be aggressive about giving up on initial packet
         self.incommand = False
         self.initialized = True
         self.cleaningup = False
         self.lastpayload = None
+        self._customkeepalives = None
         self.bmc = bmc
+        self.broken = False
         try:
             self.userid = userid.encode('utf-8')
             self.password = password.encode('utf-8')
@@ -348,14 +368,16 @@ class Session(object):
         else:
             self.kg = self.password
         self.port = port
-        if (onlogon is None):
+        if onlogon is None:
             self.async = False
             self.logonwaiters = [self._sync_login]
         else:
             self.async = True
             self.logonwaiters = [onlogon]
-        if not hasattr(Session, 'socket'):
-            self._createsocket()
+        if self.__class__.socketchecking is None:
+            self.__class__.socketchecking = threading.Lock()
+        with self.socketchecking:
+            self.socket = self._assignsocket()
         self.login()
         if not self.async:
             while not self.logged:
@@ -365,15 +387,20 @@ class Session(object):
         # since our connection has failed retries
         # deregister our keepalive facility
         Session.keepalive_sessions.pop(self, None)
-        self.logged = 0  # mark session as busted
-        # since this session is broken, remove it from the handler list.
-        # This allows constructor to create a new, functional object to
-        # replace this one
-        for sockaddr in self.allsockaddrs:
-            if sockaddr in Session.bmc_handlers:
-                del Session.bmc_handlers[sockaddr]
-        if self.sol_handler:
-            self.sol_handler({'error': 'Session Disconnected'})
+        if self.logged:
+            self.logged = 0  # mark session as busted
+            self._customkeepalives = None
+            if not self.broken:
+                self.socketpool[self.socket] -= 1
+                self.broken = True
+                # since this session is broken, remove it from the handler list
+                # This allows constructor to create a new, functional object to
+                # replace this one
+                for sockaddr in self.allsockaddrs:
+                    if sockaddr in Session.bmc_handlers:
+                        del Session.bmc_handlers[sockaddr]
+                if self.sol_handler:
+                    self.sol_handler({'error': 'Session Disconnected'})
 
     def onlogon(self, parameter):
         if 'error' in parameter:
@@ -387,11 +414,6 @@ class Session(object):
         #                 I picked 'xCAT' minus 1 so that a hexdump of packet
         #                 would show xCAT
         self.localsid = 2017673555
-
-        # NOTE(jbjohnso): for the moment, assume admin access
-        # TODO(jbjohnso): make flexible
-        self.privlevel = 4
-
         self.confalgo = 0
         self.aeskey = None
         self.integrityalgo = 0
@@ -426,24 +448,17 @@ class Session(object):
         self.sol_handler = None
         # NOTE(jbjohnso): This is the callback handler for any SOL payload
 
-    def _checksum(self, *data):  # Two's complement over the data
-        csum = sum(data)
-        csum = csum ^ 0xff
-        csum += 1
-        csum &= 0xff
-        return csum
-
     def _make_bridge_request_msg(self, channel, netfn, command):
         """This function generate message for bridge request. It is a
         part of ipmi payload.
         """
         head = [constants.IPMI_BMC_ADDRESS,
                 constants.netfn_codes['application'] << 2]
-        check_sum = self._checksum(*head)
+        check_sum = _checksum(*head)
         #NOTE(fengqian): according IPMI Figure 14-11, rqSWID is set to 81h
-        boday = [0x81, self.seqlun, constants.IPMI_SEND_MESSAGE_CMD]
+        boday = [0x81, self.seqlun, constants.IPMI_SEND_MESSAGE_CMD,
+                 0x40 | channel]
         #NOTE(fengqian): Track request
-        boday.append(0x40 | channel)
         self._add_request_entry((constants.netfn_codes['application'] + 1,
                                  self.seqlun, constants.IPMI_SEND_MESSAGE_CMD))
         return head + [check_sum] + boday
@@ -463,12 +478,11 @@ class Session(object):
         if self._lookup_request_entry(entry):
             self.request_entry.remove(entry)
 
-    def _make_ipmi_payload(self, netfn, command, bridge_request={}, data=()):
+    def _make_ipmi_payload(self, netfn, command, bridge_request=None, data=()):
         """This function generates the core ipmi payload that would be
         applicable for any channel (including KCS)
         """
         bridge_msg = []
-        payload = []
         self.expectedcmd = command
         self.expectednetfn = netfn + \
             1  # in ipmi, the response netfn is always one
@@ -502,13 +516,13 @@ class Session(object):
         header = [rsaddr, netfn << 2]
 
         reqbody = [rqaddr, self.seqlun, command] + list(data)
-        headsum = self._checksum(*header)
-        bodysum = self._checksum(*reqbody)
+        headsum = _checksum(*header)
+        bodysum = _checksum(*reqbody)
         payload = header + [headsum] + reqbody + [bodysum]
         if bridge_request:
             payload = bridge_msg + payload
             #NOTE(fengqian): For bridge request, another check sum is needed.
-            tail_csum = self._checksum(*payload[3:])
+            tail_csum = _checksum(*payload[3:])
             payload.append(tail_csum)
 
         self._add_request_entry((self.expectednetfn, self.seqlun, command))
@@ -523,8 +537,8 @@ class Session(object):
     def raw_command(self,
                     netfn,
                     command,
-                    bridge_request={},
-                    data=[],
+                    bridge_request=None,
+                    data=(),
                     retry=True,
                     delay_xmit=None):
         if not self.logged:
@@ -554,7 +568,7 @@ class Session(object):
             Session.wait_for_rsp(timeout=timeout)
         return self.lastresponse
 
-    def _send_ipmi_net_payload(self, netfn, command, data, bridge_request={},
+    def _send_ipmi_net_payload(self, netfn, command, data, bridge_request=None,
                                retry=True, delay_xmit=None):
         ipmipayload = self._make_ipmi_payload(netfn, command, bridge_request,
                                               data)
@@ -562,7 +576,7 @@ class Session(object):
         self.send_payload(payload=ipmipayload, payload_type=payload_type,
                           retry=retry, delay_xmit=delay_xmit)
 
-    def send_payload(self, payload=None, payload_type=None, retry=True,
+    def send_payload(self, payload=(), payload_type=None, retry=True,
                      delay_xmit=None, needskeepalive=False):
         """Send payload over the IPMI Session
 
@@ -573,8 +587,8 @@ class Session(object):
                                Notably, 0-length SOL packets
                                are prone to confusion.
         """
-        if payload is not None and self.lastpayload is not None:
-                             #we already have a packet outgoing, make this
+        if payload and self.lastpayload:
+                             # we already have a packet outgoing, make this
                              # a pending payload
                              # this way a simplistic BMC won't get confused
                              # and we also avoid having to do more complicated
@@ -584,7 +598,7 @@ class Session(object):
             return
         if payload_type is None:
             payload_type = self.last_payload_type
-        if payload is None:
+        if not payload:
             payload = self.lastpayload
         message = [0x6, 0, 0xff, 0x07]  # constant RMCP header for IPMI
         if retry:
@@ -596,9 +610,9 @@ class Session(object):
             payload_type |= 0b01000000
         if self.confalgo:
             payload_type |= 0b10000000
-        if (self.ipmiversion == 2.0):
+        if self.ipmiversion == 2.0:
             message.append(payload_type)
-            if (baretype == 2):
+            if baretype == 2:
                 #TODO(jbjohnso): OEM payload types
                 raise NotImplementedError("OEM Payloads")
             elif baretype not in constants.payload_types.values():
@@ -606,7 +620,7 @@ class Session(object):
                     "Unrecognized payload type %d" % baretype)
             message += struct.unpack("!4B", struct.pack("<I", self.sessionid))
         message += struct.unpack("!4B", struct.pack("<I", self.sequencenumber))
-        if (self.ipmiversion == 1.5):
+        if self.ipmiversion == 1.5:
             message += struct.unpack("!4B", struct.pack("<I", self.sessionid))
             if not self.authtype == 0:
                 message += self._ipmi15authcode(payload)
@@ -615,7 +629,7 @@ class Session(object):
             totlen = 34 + \
                 len(message)  # Guessing the ipmi spec means the whole
                                    # packet and assume no tag in old 1.5 world
-            if (totlen in (56, 84, 112, 128, 156)):
+            if totlen in (56, 84, 112, 128, 156):
                 message.append(0)  # Legacy pad as mandated by ipmi spec
         elif self.ipmiversion == 2.0:
             psize = len(payload)
@@ -734,10 +748,9 @@ class Session(object):
         self.sessionid = struct.unpack("<I", struct.pack("4B", *data[0:4]))[0]
         self.authtype = 2
         self._activate_session(data[4:])
-    '''
-    This sends the activate session payload.  We pick '1' as the requested
-    sequence number without perturbing our real sequence number
-    '''
+    # NOTE(jbjohnso):
+    # This sends the activate session payload.  We pick '1' as the requested
+    # sequence number without perturbing our real sequence number
 
     def _activate_session(self, data):
         rqdata = [2, 4] + list(data) + [1, 0, 0, 0]
@@ -811,7 +824,7 @@ class Session(object):
 
     def _get_channel_auth_cap(self):
         self.ipmicallback = self._got_channel_auth_cap
-        if (self.ipmi15only):
+        if self.ipmi15only:
             self._send_ipmi_net_payload(netfn=0x6,
                                         command=0x38,
                                         data=[0x0e, self.privlevel])
@@ -826,12 +839,13 @@ class Session(object):
         self._get_channel_auth_cap()
 
     @classmethod
-    def pulltoqueue(cls, mysocket, queue):
-        while True:
-            rdata = _io_apply(_io_recvfrom, (mysocket, 3000))
-            if rdata is None:
-                break
-            queue.append(rdata)
+    def pulltoqueue(cls, mysockets, queue):
+        for mysocket in mysockets:
+            while True:
+                rdata = _io_apply(_io_recvfrom, (mysocket, 3000))
+                if rdata is None:
+                    break
+                queue.append(rdata)
 
     @classmethod
     def wait_for_rsp(cls, timeout=None, callout=True):
@@ -891,17 +905,14 @@ class Session(object):
                 timeout = 0
         if timeout is None:
             return 0
-        rdylist = _poller(iosockets, timeout=timeout)
+        rdylist = _poller(timeout=timeout)
         if len(rdylist) > 0:
-            while _poller((cls.socket,)):  # if the somewhat lengthy
-                        # queue # processing takes long enough for packets to
-                        # come in, be eager
-                pktqueue = collections.deque([])
-                cls.pulltoqueue(cls.socket, pktqueue)
-                while len(pktqueue):
-                    (data, sockaddr) = pktqueue.popleft()
-                    cls._route_ipmiresponse(sockaddr, data)
-                    cls.pulltoqueue(cls.socket, pktqueue)
+            pktqueue = collections.deque([])
+            cls.pulltoqueue(iosockets, pktqueue)
+            while len(pktqueue):
+                (data, sockaddr) = pktqueue.popleft()
+                cls._route_ipmiresponse(sockaddr, data)
+                cls.pulltoqueue(iosockets, pktqueue)
         sessionstodel = []
         sessionstokeepalive = []
         for session, parms in cls.keepalive_sessions.iteritems():
@@ -920,17 +931,54 @@ class Session(object):
                     session)  # defer deletion until after loop
                                               # to avoid confusing the for loop
         for session in sessionstodel:
-            cls.pending -= 1
             cls.waiting_sessions.pop(session, None)
             session._timedout()
         return len(cls.waiting_sessions)
 
+    def register_keepalive(self, cmd, callback):
+        '''Register  custom keepalive IPMI command
+
+        This is mostly intended for use by the console code.
+        calling code would have an easier time just scheduling in their
+        own threading scheme.  Such a behavior would naturally cause
+        the default keepalive to not occur anyway if the calling code
+        is at least as aggressive about timing as pyghmi
+        :param cmd: A dict of arguments to be passed into raw_command
+        :param callback: A function to be called with results of the keepalive
+
+        :returns: value to identify registration for unregister_keepalive
+        '''
+        regid = random.random()
+        if self._customkeepalives is None:
+            self._customkeepalives = {regid: (cmd, callback)}
+        else:
+            while regid in self._customkeepalives:
+                regid = random.random()
+            self._customkeepalives[regid] = (cmd, callback)
+        return regid
+
+    def unregister_keepalive(self, regid):
+        try:
+            del self._customkeepalives[regid]
+        except KeyError:
+            pass
+
     def _keepalive(self):
         """Performs a keepalive to avoid idle disconnect
         """
-        if self.incommand:  # if currently in command, no cause to keepalive
-            return
-        self.raw_command(netfn=6, command=1)
+        try:
+            if not self._customkeepalives:
+                if self.incommand:
+                    # if currently in command, no cause to keepalive
+                    return
+                self.raw_command(netfn=6, command=1)
+            else:
+                kaids = list(self._customkeepalives.keys())
+                for keepalive in kaids:
+                    cmd, callback = self._customkeepalives[keepalive]
+                    callback(self.raw_command(**cmd))
+        except exc.IpmiException:
+            self._mark_broken()
 
     @classmethod
     def _route_ipmiresponse(cls, sockaddr, data):
@@ -939,7 +987,6 @@ class Session(object):
         try:
             cls.bmc_handlers[sockaddr]._handle_ipmi_packet(data,
                                                            sockaddr=sockaddr)
-            cls.pending -= 1
         except KeyError:
             pass
 
@@ -1069,7 +1116,10 @@ class Session(object):
             self.onlogon({'error': errstr})
             return -9
         self.allowedpriv = data[2]
-        # TODO(jbjohnso): enable lower priv access (e.g. operator/user)
+        # NOTE(jbjohnso): At this point, the BMC has no idea about what user
+        # shall be used.  As such, the allowedpriv field is actually
+        # not particularly useful.  got_rakp2 is a good place to
+        # gracefully detect and downgrade privilege for retry
         localsid = struct.unpack("<I", struct.pack("4B", *data[4:8]))[0]
         if self.localsid != localsid:
             return -9
@@ -1103,6 +1153,12 @@ class Session(object):
         if data[0] != self.rmcptag:  # ignore mismatched tags for retry logic
             return -9
         if data[1] != 0:  # if not successful, consider next move
+            if data[1] == 9 and self.privlevel == 4:
+                # Here the situation is likely that the peer didn't want
+                # us to use Operator.  Degrade to operator and try again
+                self.privlevel = 3
+                self.login()
+                return
             if data[1] == 2:  # invalid sessionid 99% of the time means a retry
                              # scenario invalidated an in-flight transaction
                 return
@@ -1198,10 +1254,7 @@ class Session(object):
         self.lastpayload = None
         self._req_priv_level()
 
-    '''
-    Internal function to parse IPMI nugget once extracted from its framing
-    '''
-
+    # Internal function to parse IPMI nugget once extracted from its framing
     def _parse_ipmi_payload(self, payload):
         # For now, skip the checksums since we are in LAN only,
         # TODO(jbjohnso): if implementing other channels, add checksum checks
@@ -1311,8 +1364,6 @@ class Session(object):
             Session.wait_for_rsp(timeout=0, callout=False)  # take opportunity
                                                  # to drain the socket queue if
                                                  # applicable
-            while Session.pending > Session.maxpending:
-                Session.wait_for_rsp()
         if self.sequencenumber:  # seq number of zero will be left alone, it is
                                 # special, otherwise increment
             self.sequencenumber += 1
@@ -1321,14 +1372,13 @@ class Session(object):
             Session.waiting_sessions[self]['ipmisession'] = self
             Session.waiting_sessions[self]['timeout'] = self.timeout + \
                 _monotonic_time()
-            Session.pending += 1
         if delay_xmit is not None:
             Session.waiting_sessions[self]['timeout'] = delay_xmit + \
                 _monotonic_time()
             return  # skip transmit, let retry timer do it's thing
         if self.sockaddr:
             _io_apply(_io_sendto,
-                      (Session.socket, self.netpacket, self.sockaddr))
+                      (self.socket, self.netpacket, self.sockaddr))
         else:  # he have not yet picked a working sockaddr for this connection,
               # try all the candidates that getaddrinfo provides
             self.allsockaddrs = []
@@ -1338,13 +1388,13 @@ class Session(object):
                                               0,
                                               socket.SOCK_DGRAM):
                     sockaddr = res[4]
-                    if (res[0] == socket.AF_INET):  # convert the sockaddr
+                    if res[0] == socket.AF_INET:  # convert the sockaddr
                                                     # to AF_INET6
                         newhost = '::ffff:' + sockaddr[0]
                         sockaddr = (newhost, sockaddr[1], 0, 0)
                     self.allsockaddrs.append(sockaddr)
                     Session.bmc_handlers[sockaddr] = self
-                    _io_apply(_io_sendto, (Session.socket,
+                    _io_apply(_io_sendto, (self.socket,
                               self.netpacket, sockaddr))
             except socket.gaierror:
                 raise exc.IpmiException(
@@ -1360,8 +1410,12 @@ class Session(object):
                          data=struct.unpack("4B",
                                             struct.pack("I", self.sessionid)),
                          retry=False)
+        # stop trying for a keepalive,
+        Session.keepalive_sessions.pop(self, None)
         self.logged = 0
+        self._customkeepalives = None
         self.nowait = False
+        self.socketpool[self.socket] -= 1
         return {'success': True}
 
 
